/*
 * This IPs is the RAM memory for the Atmel MEGA CPU implementation.
 * 
 * Copyright (C) 2017  Iulian Gheorghiu  (morgoth@devboard.tech)
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

`timescale 1ns / 1ps
 
 module mega_ram  #(
	parameter ADDR_BUS_WIDTH = 13,  /* < in address lines */
	parameter ADDR_RAM_DEPTH = 2 ** ADDR_BUS_WIDTH,  /* < in address lines */
	parameter DATA_BUS_WIDTH = 8
) (
	input clk,
	input cs,
	input we,
	input re,
	input [ADDR_BUS_WIDTH-1:0] a,
	input [DATA_BUS_WIDTH - 1:0] d_in,
	output [DATA_BUS_WIDTH - 1:0] d_out
);

generate
/* MEM_MODE == "SRAM" */

wire [15:0]d_out_w;
wire [DATA_BUS_WIDTH - 1:0]d_out_w_b;
reg [DATA_BUS_WIDTH - 1:0]d_out_tmp;
reg [ADDR_BUS_WIDTH-1:0] addr;


always @ (*)
begin
	addr = a;
end

reg delayed_a0;

always @ (posedge clk) delayed_a0 = addr[0];

wire [13:0]ram_addr = DATA_BUS_WIDTH == 8 ? addr[ADDR_BUS_WIDTH-1 : 1] : addr;

	
EG_LOGIC_BRAM #( .DATA_WIDTH_A(16),
				.ADDR_WIDTH_A(14),
				.DATA_DEPTH_A(16384),
				.DATA_WIDTH_B(16),
				.ADDR_WIDTH_B(14),
				.DATA_DEPTH_B(16384),
				.BYTE_ENABLE(8),
				.BYTE_A(2),
				.BYTE_B(2),
				.MODE("SP"),
				.REGMODE_A("OUTREG"),
				.WRITEMODE_A("NORMAL"),
				.RESETMODE("SYNC"),
				.IMPLEMENT("32K"),
				.DEBUGGABLE("NO"),
				.PACKABLE("NO"),
				.INIT_FILE("NONE"),
				.FILL_ALL("NONE")				
) inst (
				.dia(DATA_BUS_WIDTH == 8 ? {d_in[7:0], d_in[7:0]} : d_in),
				.dib({16{1'b0}}),
				.addra(ram_addr),
				.addrb({14{1'b0}}),
				.cea(1'b1),
				.ceb(1'b0),
				.ocea(1'b1),
				.oceb(1'b0),
				.clka(clk),
				.clkb(1'b0),				
				.bea(DATA_BUS_WIDTH == 8 ? (addr[0] ? 2'b10 : 2'b01) : 2'b11),				
				.beb(2'b00),
				.wea(we),
				.web(1'b0),
				.rsta(1'b0),
				.rstb(1'b0),
				.doa(d_out_w),
				.dob());


assign d_out_w_b = DATA_BUS_WIDTH == 8 ? (delayed_a0 ? d_out_w[15:8] : d_out_w[7:0]) : d_out_w;

always @ (*) begin 	d_out_tmp = d_out_w_b;	end


assign d_out = (cs & re) ? d_out_tmp : 'b00;


endgenerate
endmodule
