/*
 * This IP is the core registers implementation.
 * 
 * Copyright (C) 2018  Iulian Gheorghiu (morgoth.creator@gmail.com)
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

`timescale 1ns / 1ps

`include "mega-def.v"

module mega_regs #
	(
	parameter PLATFORM = "XILINX",		// XILINX	ANLOGIC_EG4S20
	parameter REGISTERED = "FALSE"
	)(
	input rst,
	input clk,
	input [4:0]rs1a,
	output reg[15:0]rs1,
	input rs1m,
	input [4:0]rs2a,
	output reg[15:0]rs2,
	input rs2m,
	input [4:0]rda,
	input [15:0]rd,
	input rdw,
	input rdm
	);

generate
begin/* Anlogic TD does not know how to implement distributed RAM with true three ports, so I implement him from individual distributed RAM cells. */
	
wire [7:0] REGLD_out;
wire [7:0] REGHD_out;
wire [7:0] REGLR_out;
wire [7:0] REGHR_out;
wire write_to_L = rdw & ((!rdm & !rda[0]) || rdm);
wire write_to_H = rdw & ((!rdm & rda[0]) || rdm);
wire write_to_HL = rdw & rdm;

/* REG_L_D_0_7 */
EG_LOGIC_DRAM #(
 			.INIT_FILE("NONE"),
			.DATA_WIDTH_W(8),
			.ADDR_WIDTH_W(4),
			.DATA_DEPTH_W(16),
			.DATA_WIDTH_R(8),
			.ADDR_WIDTH_R(4),
			.DATA_DEPTH_R(16)
	) REG_L_D_0_7 (
			.di(rd[7:0]),
			.waddr(write_to_HL ? rda[3:0] : rda[4:1]),
			.wclk(clk),
			.we(write_to_L | write_to_HL),
			.do(REGLD_out[7:0]),
			.raddr(rs1m ? rs2a[3:0] : rs1a[4:1]));

/* REG_H_D_0_7 */
EG_LOGIC_DRAM #(
 			.INIT_FILE("NONE"),
			.DATA_WIDTH_W(8),
			.ADDR_WIDTH_W(4),
			.DATA_DEPTH_W(16),
			.DATA_WIDTH_R(8),
			.ADDR_WIDTH_R(4),
			.DATA_DEPTH_R(16)
	) REG_H_D_0_7 (
			.di(write_to_HL ? rd[15:8] : rd[7:0]),
			.waddr(write_to_HL ? rda[3:0] : rda[4:1]),
			.wclk(clk),
			.we(write_to_H | write_to_HL),
			.do(REGHD_out[7:0]),
			.raddr(rs1m ? rs2a[3:0] : rs1a[4:1]));


/* REG_L_R_0_7 */
EG_LOGIC_DRAM #(
 			.INIT_FILE("NONE"),
			.DATA_WIDTH_W(8),
			.ADDR_WIDTH_W(4),
			.DATA_DEPTH_W(16),
			.DATA_WIDTH_R(8),
			.ADDR_WIDTH_R(4),
			.DATA_DEPTH_R(16)
	) REG_L_R_0_7 (
			.di(rd[7:0]),
			.waddr(write_to_HL ? rda[3:0]:rda[4:1]),
			.wclk(clk),
			.we(write_to_L | write_to_HL),
			.do(REGLR_out[7:0]),
			.raddr(rs2m ? rs2a[3:0]:rs2a[4:1]));

	
/* REG_H_R_0_7 */
EG_LOGIC_DRAM #(
 			.INIT_FILE("NONE"),
			.DATA_WIDTH_W(8),
			.ADDR_WIDTH_W(4),
			.DATA_DEPTH_W(16),
			.DATA_WIDTH_R(8),
			.ADDR_WIDTH_R(4),
			.DATA_DEPTH_R(16)
	) REG_H_R_0_7 (
			.di(write_to_HL ? rd[15:8] : rd[7:0]),
			.waddr(write_to_HL ? rda[3:0]:rda[4:1]),
			.wclk(clk),
			.we(write_to_H | write_to_HL),
			.do(REGHR_out[7:0]),
			.raddr(rs2m ? rs2a[3:0]:rs2a[4:1]));	
	

always @ *
	begin
		case({rs1m, rs1a[0]})
			2'b00: 		rs1 <= {8'h00, REGLD_out};
			2'b01: 		rs1 <= {8'h00, REGHD_out};
			default : 	rs1 <= {REGHD_out, REGLD_out};
		endcase
		
		case({rs2m, rs2a[0]})
			2'b00: 		rs2 <= {8'h00, REGLR_out};
			2'b01: 		rs2 <= {8'h00, REGHR_out};
			default : 	rs2 <= {REGHR_out, REGLR_out};
		endcase
	end
end/* PLATFORM != "ANLOGIC_EG4S20" */	
endgenerate
	
endmodule
