/*
 * This IP is the ROM memory for the Atmel MEGA CPU implementation.
 * 
 * Copyright (C) 2017  Iulian Gheorghiu  (morgoth@devboard.tech)
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

`timescale 1ns / 1ps
 
module mega_rom  #(
	parameter ADDR_ROM_BUS_WIDTH = 14, /* < in address lines */
	parameter ROM_PATH = ""
) (
	input clk,
	input [ADDR_ROM_BUS_WIDTH-1:0] a,
	input cs,
	output [15:0]d
);

reg[15:0]d_int;
wire[15:0]dw_int;
generate
begin

EG_LOGIC_BRAM 
#(
	.DATA_WIDTH_A(16),
	.ADDR_WIDTH_A(ADDR_ROM_BUS_WIDTH),
	.DATA_DEPTH_A(2 ** ADDR_ROM_BUS_WIDTH),
	.DATA_WIDTH_B(16),
	.ADDR_WIDTH_B(ADDR_ROM_BUS_WIDTH),
	.DATA_DEPTH_B(2 ** ADDR_ROM_BUS_WIDTH),
	.MODE("SP"),
	.REGMODE_A("NOREG"),
	.IMPLEMENT("32K"),
	.DEBUGGABLE("NO"),
	.PACKABLE("NO"),
	.INIT_FILE(ROM_PATH),
	.FILL_ALL("NONE")
) rom_inst (
	.dia({16{1'b0}}),
	.dib({16{1'b0}}),
	.addra(a),
	.addrb({ADDR_ROM_BUS_WIDTH{1'b0}}),
	.cea(1'b1),
	.ceb(1'b0),
	.ocea(1'b0),
	.oceb(1'b0),
	.clka(clk),
	.clkb(1'b0),
	.wea(1'b0),
	.web(1'b0),
	.bea(1'b0),
	.beb(1'b0),
	.rsta(1'b0),
	.rstb(1'b0),
	.doa(dw_int),
	.dob()
);

assign d = cs ? dw_int : 16'h0000;

end
endgenerate
endmodule


