/*
 * This IP is the top module for a simple template implementation.
 * 
 * Copyright (C) 2020  Iulian Gheorghiu (morgoth@devboard.tech)
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

`timescale 1ns / 1ps


`define PLATFORM				"ANLOGIC_EG4S20"	//	"iCE40UP"
`define FLASH_ROM_FILE_NAME		"../prog/test5_soft_core.mif"

module top (
	input 	i_CLK24,
	input	i_USER_N,
	output 	RGB0, 
	output 	RGB1, 
	output 	RGB2,	
		
	output [15:0] o_DATA16,
 
	input 	BTN_UP,
	input 	BTN_DN,
	input 	BTN_BACK,
	input 	BTN_OK,
	input 	BTN_INTERRUPT
	);

wire pll_lock;
wire sys_clk, pll_clk;
/* We use the 'pll_locked' signal from the PLL to release the CORE and IO RESET line. */
wire sys_rst = ~(i_USER_N & pll_lock);
wire [2:0]LED;

/* PLL Instance */
PLL_DEV PLL_inst(
	.refclk (i_CLK24), 
	.reset (~i_USER_N), 
	.extlock (pll_lock), 
	.clk0_out (pll_clk), 		// 192.00000 MHZ
	.clk1_out (sys_clk) 		//  16.00000 MHZ
);
/* !PLL Instance */
 

/* CORE Instance */
core # (
	.PLATFORM(`PLATFORM),
	.ROM_PATH(`FLASH_ROM_FILE_NAME)
)core_instance(
	.core_rst(sys_rst),
	.core_clk(sys_clk),
    .RGB_LED(LED),	
	.DEBUG_DAT (o_DATA16),
	.BTN_UP(BTN_UP),
	.BTN_DN(BTN_DN),
	.BTN_BACK(BTN_BACK),
	.BTN_OK(BTN_OK),
	.BTN_INTERRUPT(BTN_INTERRUPT)
);
/* !CORE Instance */

assign RGB0 = ~LED[0];
assign RGB1 = ~LED[1];
assign RGB2 = ~LED[2];


endmodule